﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Security.Cryptography;
using System.Diagnostics;

using BlakeSharp;

namespace BlakeSharpDemo
{
	public static class Program
	{
		private enum OpMode
		{
			File,
			AnsiString,
			TestVectorFile
		}

		private const StringComparison StrCaseIgnoreCmp =
			StringComparison.OrdinalIgnoreCase;

		public static void Main(string[] args)
		{
			try { Run(args); }
			catch(Exception ex) { Console.WriteLine(ex.Message); }
		}

		private static void Run(string[] args)
		{
			OpMode opMode = OpMode.File;

			for(int i = 0; i < args.Length; ++i)
			{
				string strArg = args[i];
				if(string.IsNullOrEmpty(strArg)) continue;

				if(strArg.Equals("-f", StrCaseIgnoreCmp))
					opMode = OpMode.File;
				else if(strArg.Equals("-s", StrCaseIgnoreCmp))
					opMode = OpMode.AnsiString;
				else if(strArg.Equals("-t", StrCaseIgnoreCmp))
					SelfTest();
				else if(strArg.Equals("-tf", StrCaseIgnoreCmp))
					opMode = OpMode.TestVectorFile;
				else if(strArg.Equals("-b256", StrCaseIgnoreCmp))
					Benchmark(new Blake256());
				else if(strArg.Equals("-b512", StrCaseIgnoreCmp))
					Benchmark(new Blake512());
				else
				{
					if(opMode == OpMode.File)
						HashFile(strArg);
					else if(opMode == OpMode.AnsiString)
						HashString(strArg, Encoding.Default);
					else if(opMode == OpMode.TestVectorFile)
						EvalTestVectorFile(strArg);
					else { Debug.Assert(false); }
				}
			}
		}

		private static void HashFile(string strFilePath)
		{
			Console.WriteLine(@"File '" + strFilePath + @"'");

			try
			{
				FileStream fsIn = new FileStream(strFilePath, FileMode.Open,
					FileAccess.Read, FileShare.Read);
				Blake256 blake256 = new Blake256();
				byte[] pbHash = blake256.ComputeHash(fsIn);
				fsIn.Close();
				Console.WriteLine("BLAKE-256:");
				Console.WriteLine(MemUtil.ByteArrayToHexString(pbHash));

				fsIn = new FileStream(strFilePath, FileMode.Open,
					FileAccess.Read, FileShare.Read);
				Blake512 blake512 = new Blake512();
				pbHash = blake512.ComputeHash(fsIn);
				fsIn.Close();
				Console.WriteLine("BLAKE-512:");
				Console.WriteLine(MemUtil.ByteArrayToHexString(pbHash));

				Console.WriteLine();
			}
			catch(Exception ex) { Console.WriteLine(ex.Message); }
		}

		private static void HashString(string strText, Encoding enc)
		{
			if(enc == null) enc = Encoding.Default;

			Console.WriteLine(@"Text '" + strText + @"' (encoding " +
				enc.WebName + ")");

			byte[] pbData = enc.GetBytes(strText);

			Blake256 blake256 = new Blake256();
			byte[] pbHash256 = blake256.ComputeHash(pbData);
			Console.WriteLine("BLAKE-256:");
			Console.WriteLine(MemUtil.ByteArrayToHexString(pbHash256));

			Blake512 blake512 = new Blake512();
			byte[] pbHash512 = blake512.ComputeHash(pbData);
			Console.WriteLine("BLAKE-512:");
			Console.WriteLine(MemUtil.ByteArrayToHexString(pbHash512));

			Console.WriteLine();
		}

		private static void ThrowIfNotEqual(byte[] pb1, byte[] pb2, string strMsg)
		{
			if(!MemUtil.ArraysEqual(pb1, pb2)) throw new Exception(strMsg);
		}

		private static void SelfTest()
		{
			try
			{
				SelfTestPriv();
				Console.WriteLine("All self-tests passed successfully!");
			}
			catch(Exception ex)
			{
				Console.WriteLine("Self-test failed: " + ex.Message + "!");
			}
		}

		private static void SelfTestPriv()
		{
			Blake512 blake512 = new Blake512();

			byte[] pbData = new byte[1] { 0 };
			byte[] pbExpc = new byte[64] {
				0x97, 0x96, 0x15, 0x87, 0xF6, 0xD9, 0x70, 0xFA,
				0xBA, 0x6D, 0x24, 0x78, 0x04, 0x5D, 0xE6, 0xD1,
				0xFA, 0xBD, 0x09, 0xB6, 0x1A, 0xE5, 0x09, 0x32,
				0x05, 0x4D, 0x52, 0xBC, 0x29, 0xD3, 0x1B, 0xE4,
				0xFF, 0x91, 0x02, 0xB9, 0xF6, 0x9E, 0x2B, 0xBD,
				0xB8, 0x3B, 0xE1, 0x3D, 0x4B, 0x9C, 0x06, 0x09,
				0x1E, 0x5F, 0xA0, 0xB4, 0x8B, 0xD0, 0x81, 0xB6,
				0x34, 0x05, 0x8B, 0xE0, 0xEC, 0x49, 0xBE, 0xB3
			};
			byte[] pbHash = blake512.ComputeHash(pbData);
			ThrowIfNotEqual(pbHash, pbExpc, "ST-1");

			pbData = new byte[0];
			pbExpc = new byte[64] {
				0xA8, 0xCF, 0xBB, 0xD7, 0x37, 0x26, 0x06, 0x2D,
				0xF0, 0xC6, 0x86, 0x4D, 0xDA, 0x65, 0xDE, 0xFE,
				0x58, 0xEF, 0x0C, 0xC5, 0x2A, 0x56, 0x25, 0x09,
				0x0F, 0xA1, 0x76, 0x01, 0xE1, 0xEE, 0xCD, 0x1B,
				0x62, 0x8E, 0x94, 0xF3, 0x96, 0xAE, 0x40, 0x2A,
				0x00, 0xAC, 0xC9, 0xEA, 0xB7, 0x7B, 0x4D, 0x4C,
				0x2E, 0x85, 0x2A, 0xAA, 0xA2, 0x5A, 0x63, 0x6D,
				0x80, 0xAF, 0x3F, 0xC7, 0x91, 0x3E, 0xF5, 0xB8
			};
			pbHash = blake512.ComputeHash(pbData);
			ThrowIfNotEqual(pbHash, pbExpc, "ST-2");
		}

		private static void EvalTestVectorFile(string strFilePath)
		{
			Blake256 blake256 = new Blake256();
			Blake512 blake512 = new Blake512();
			uint uSucceeded = 0, uFailed = 0;

			try
			{
				string[] v = File.ReadAllLines(strFilePath);
				if((v == null) || (v.Length == 0)) throw new FileNotFoundException();

				byte[] pbMsg = new byte[0];
				ulong uRepeat = 1;
				long lLength = -1;
				for(int i = 0; i < v.Length; ++i)
				{
					string strLine = v[i].Trim();
					if(strLine.Length == 0) continue;

					if(strLine.StartsWith(@"#", StrCaseIgnoreCmp)) { } // Comment
					else if(strLine.StartsWith("Repeat = ", StrCaseIgnoreCmp))
						uRepeat = ulong.Parse(strLine.Substring(9));
					else if(strLine.StartsWith("Len = ", StrCaseIgnoreCmp))
						lLength = long.Parse(strLine.Substring(6));
					else if(strLine.StartsWith("Text = ", StrCaseIgnoreCmp))
						pbMsg = Encoding.Default.GetBytes(strLine.Substring(7));
					else if(strLine.StartsWith("Msg = ", StrCaseIgnoreCmp))
						pbMsg = MemUtil.HexStringToByteArray(strLine.Substring(6));
					else if(strLine.StartsWith("MD = ", StrCaseIgnoreCmp))
					{
						if((lLength < 0) || (((lLength % 8) == 0) && (lLength > 0)))
						{
							string strExpc = strLine.Substring(5);
							byte[] pbExpc = MemUtil.HexStringToByteArray(strExpc);

							if(uRepeat > 1)
							{
								MemoryStream ms = new MemoryStream();
								for(ulong u = 0; u < uRepeat; ++u)
									ms.Write(pbMsg, 0, pbMsg.Length);
								pbMsg = ms.ToArray();
								ms.Close();
							}

							byte[] pbComp = new byte[0];
							if(strExpc.Length == (32 * 2))
								pbComp = blake256.ComputeHash(pbMsg);
							else if(strExpc.Length == (64 * 2))
								pbComp = blake512.ComputeHash(pbMsg);

							if(MemUtil.ArraysEqual(pbExpc, pbComp))
								// Console.WriteLine("Validated test vector successfully (hash " +
								//	strExpc.Substring(0, 8) + "...).");
								++uSucceeded;
							else
							{
								Console.WriteLine("Test vector failed (hash " +
									strExpc.Substring(0, 8) + "...).");
								++uFailed;
							}
						}

						pbMsg = new byte[0];
						uRepeat = 1;
						lLength = -1;
					}
					else { Debug.Assert(false); }
				}
			}
			catch(Exception ex) { Console.WriteLine(ex.Message); }

			Console.WriteLine("Summary: " + uSucceeded.ToString() +
				" test vectors validated successfully, " + uFailed.ToString() +
				" errors.");
		}

		private static void Benchmark(HashAlgorithm ha)
		{
			Random r = new Random();
			byte[] pbData = new byte[1024 * 1024 * 4];
			r.NextBytes(pbData);

			ha.Initialize();

			bool bFirst = true;
			ulong uBytesHashed = 0;
			int iStartTicks = Environment.TickCount;
			int iLastPrintTicks = iStartTicks;

			while(true)
			{
				ha.TransformBlock(pbData, 0, pbData.Length, pbData, 0);
				uBytesHashed += (ulong)pbData.Length;

				int iCurTicks = Environment.TickCount;
				if((iCurTicks - iLastPrintTicks) >= 5000)
				{
					if(!bFirst) Console.Write(", ");
					bFirst = false;

					double d = (((double)uBytesHashed / (1024.0D * 1024.0D)) /
						((double)(iCurTicks - iStartTicks) / 1000.0D));
					Console.Write(d.ToString("F2") + " MB/s");

					iLastPrintTicks = iCurTicks;
				}
			}
		}
	}
}
